/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#ifndef __LIS_MATRIX_IO_H__
#define __LIS_MATRIX_IO_H__

#define BUFSIZE				1024
#define MM_BANNER			"%%MatrixMarket"
#define MM_MTX				"matrix"
#define MM_VEC				"vector"
#define MM_FMT				"coordinate"
#define MM_TYPE_REAL		"real"
#define MM_TYPE_GENERAL		"general"
#define MM_TYPE_SYMM		"symmetric"
#define MM_REAL				0
#define MM_GENERAL			0
#define MM_SYMM				1

#define LISBanner			"#LIS"
#define ITBLBanner			"#ITBL"

#define LIS_INPUT_N				0
#define LIS_INPUT_NNZ			1
#define LIS_INPUT_NDZ			2
#define LIS_INPUT_NND			3
#define LIS_INPUT_MAXNZR			4
#define LIS_INPUT_NR				5
#define LIS_INPUT_NC				6
#define LIS_INPUT_BNNZ			7
#define LIS_INPUT_BNR			8
#define LIS_INPUT_BNC			9
#define LIS_INPUT_NNZL			10
#define LIS_INPUT_NNZU			11

#define LIS_INPUT_ROW			0
#define LIS_INPUT_COL			1
#define LIS_INPUT_PTR			2
#define LIS_INPUT_INDEX			3
#define LIS_INPUT_BPTR			4
#define LIS_INPUT_BINDEX			5
#define LIS_INPUT_VALUE			6
#define LIS_INPUT_SOLUTION		7
#define LIS_INPUT_RHS			8
#define LIS_INPUT_DIAG			9
#define LIS_INPUT_LPTR			10
#define LIS_INPUT_LCOL			11
#define LIS_INPUT_LINDEX			12
#define LIS_INPUT_LVALUE			13
#define LIS_INPUT_UPTR			14
#define LIS_INPUT_UCOL			15
#define LIS_INPUT_UINDEX			16
#define LIS_INPUT_UVALUE			17

typedef struct
{
	char		filename[1024];
	int			fileformat;
	int			matrix_type_in;
	int			matrix_type_out;
	int			is_matrix_split;
	int			is_array_zero_cut;
	int			is_solution;
	int			is_rhs;
	int			array_base_ptr;
	int			array_base_index;
	int			precision;
	int			param_len;
	int			array_len;
	int			param_order[9];
	int			array_order[9];
} LIS_MATRIX_INPUT_OPTION;

#ifdef __cplusplus
extern "C"
{
#endif

	/****************************/
	/* Matrix INPUT             */
	/****************************/
	extern int lis_input_mm(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int matrix_type, ...);
		extern int lis_input_mm_crs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int matrix_type,int bnr, int bnc, int *row, int *col);
		extern int lis_input_mm_ccs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
		extern int lis_input_mm_msr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
		extern int lis_input_mm_dia(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
		extern int lis_input_mm_ell(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
		extern int lis_input_mm_jds(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
		extern int lis_input_mm_dns(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
		extern int lis_input_mm_coo(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm);
	extern int lis_input_mmm(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, FILE *file, LIS_Comm comm, int matrix_type, ...);
	extern int lis_input_free(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, LIS_MATRIX_INPUT_OPTION option, FILE *file, LIS_Comm comm, int matrix_type, ...);
	extern int lis_input_lis(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, char *filename, FILE *file, LIS_Comm comm, int matrix_type, ...);
		extern int lis_input_lis_crs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_ccs(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_msr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_dia(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_ell(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_jds(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_bsr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_bsc(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_vbr(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_dns(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
		extern int lis_input_lis_coo(LIS_MATRIX *A, LIS_VECTOR *b, LIS_VECTOR *x, int mode, FILE *file, LIS_Comm comm);
	/****************************/
	/* Matrix OUTPUT            */
	/****************************/
	extern int lis_output_mm_crs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_ccs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_msr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_dia(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_ell(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_jds(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_bsr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_bsc(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_vbr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_coo(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_mm_dns(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x,  char *path);
	extern int lis_output_lis_crs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_ccs(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_msr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_dia(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_ell(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_jds(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_bsr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_bsc(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_vbr(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_dns(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);
	extern int lis_output_lis_coo(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, int mode, char *path);

	/****************************/
	/* Vector OUTPUT            */
	/****************************/
	extern int lis_output_vector_plain(LIS_VECTOR v, char *path);
	extern int lis_output_vector_mm(LIS_VECTOR v, char *path);
	extern int lis_output_vector_lis_ascii(LIS_VECTOR v, char *path);
	/****************************/
	/* Vector INPUT             */
	/****************************/
	extern int lis_input_vector_plain(LIS_VECTOR *v, FILE *file, LIS_Comm comm);
	extern int lis_input_vector_mm(LIS_VECTOR *v, FILE *file, LIS_Comm comm);
	extern int lis_input_vector_lis(LIS_VECTOR *v, char *filename, FILE *file, LIS_Comm comm);
	extern int lis_input_vector_lis_ascii(LIS_VECTOR *v, FILE *file, LIS_Comm comm);

	extern int lis_fscan_int(int n, FILE *file, int val[], int origin);
	extern int lis_fscan_double(int n, FILE *file, double val[]);
	extern int lis_bswap_int(int n, int *buf);
	extern int lis_bswap_double(int n, double *buf);	
	extern int lis_bswap_size_t(int n, size_t *buf);


#ifdef __cplusplus
}
#endif
#endif
