/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#include "lis_matvec.h"

LIS_MATVEC_CRS lis_matvec_crs_xxx[5][8] = { 
	{
		lis_matvec_crs   , lis_matvec_crs_u2,lis_matvec_crs_u3 ,lis_matvec_crs_u4,
		lis_matvec_crs_u5, lis_matvec_crs_u6,lis_matvec_crs_u7 ,lis_matvec_crs_u8,
	},
	{
		lis_matvec_crs_p1,lis_matvec_crs_p2 ,lis_matvec_crs_p3 ,lis_matvec_crs_p4,
		lis_matvec_crs_p5,lis_matvec_crs_p6 ,lis_matvec_crs_p7 ,lis_matvec_crs_p8,
	},
#if defined(USE_SSE2)
	{
		lis_matvec_crs_u2_sse2, lis_matvec_crs_u2_sse2,lis_matvec_crs_u2_sse2 ,lis_matvec_crs_u4_sse2,
		lis_matvec_crs_u4_sse2, lis_matvec_crs_u4_sse2,lis_matvec_crs_u4_sse2 ,lis_matvec_crs_u4_sse2,
	},
#else
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
#endif
	{
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	},
	{
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	}
};

/************************/
/* Unrolling            */
/************************/
void lis_matvec_crs_u2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t0,t1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = 0.0;
		for(j=js+1;j<je;j+=2)
		{
			j0  = jj[0]; j1  = jj[1];

			x0  = x[j0]; x1  = x[j1];

			v0  = vv[0]; v1  = vv[1];

			t0 += v0 * x0; t1 += v1 * x1;

			jj += 2; vv += 2;
		}
		for(j-=1;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1;
	}
}

void lis_matvec_crs_u3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2;
	LIS_SCALAR	t0,t1,t2;
	LIS_SCALAR	x0,x1,x2;
	LIS_SCALAR	v0,v1,v2;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = 0.0;
		for(j=js+2;j<je;j+=3)
		{
			j0  = jj[0]; j1  = jj[1]; j2  = jj[2];

			x0  = x[j0]; x1  = x[j1]; x2  = x[j2];

			v0  = vv[0]; v1  = vv[1]; v2  = vv[2];

			t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2;

			jj += 3; vv += 3;
		}
		for(j-=2;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2;
	}
}

void lis_matvec_crs_u4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	t0,t1,t2,t3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = 0.0;
		for(j=js+3;j<je;j+=4)
		{
			j0  = jj[0]; j1  = jj[1]; j2  = jj[2]; j3  = jj[3];

			x0  = x[j0]; x1  = x[j1]; x2  = x[j2]; x3  = x[j3];

			v0  = vv[0]; v1  = vv[1]; v2  = vv[2]; v3  = vv[3];

			t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;

			jj += 4; vv += 4;
		}
		for(j-=3;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3;
	}
}

void lis_matvec_crs_u5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	t0,t1,t2,t3,t4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = 0.0;
		for(j=js+4;j<je;j+=5)
		{
			j0  = jj[0]; j1  = jj[1]; j2  = jj[2]; j3  = jj[3];
			j4  = jj[4];

			x0  = x[j0]; x1  = x[j1]; x2  = x[j2]; x3  = x[j3];
			x4  = x[j4];

			v0  = vv[0]; v1  = vv[1]; v2  = vv[2]; v3  = vv[3];
			v4  = vv[4];

			t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
			t4 += v4 * x4;

			jj += 5; vv += 5;
		}
		for(j-=4;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4;
	}
}

void lis_matvec_crs_u6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = t5 = 0.0;
		for(j=js+5;j<je;j+=6)
		{
			j0  = jj[0]; j1  = jj[1]; j2  = jj[2]; j3  = jj[3];
			j4  = jj[4]; j5  = jj[5];

			x0  = x[j0]; x1  = x[j1]; x2  = x[j2]; x3  = x[j3];
			x4  = x[j4]; x5  = x[j5];

			v0  = vv[0]; v1  = vv[1]; v2  = vv[2]; v3  = vv[3];
			v4  = vv[4]; v5  = vv[5];

			t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
			t4 += v4 * x4; t5 += v5 * x5;

			jj += 6; vv += 6;
		}
		for(j-=5;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4+t5;
	}
}

void lis_matvec_crs_u7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5,t6;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = t5 = t6 = 0.0;
		for(j=js+6;j<je;j+=7)
		{
			j0  = jj[0]; j1  = jj[1]; j2  = jj[2]; j3  = jj[3];
			j4  = jj[4]; j5  = jj[5]; j6  = jj[6];

			x0  = x[j0]; x1  = x[j1]; x2  = x[j2]; x3  = x[j3];
			x4  = x[j4]; x5  = x[j5]; x6  = x[j6];

			v0  = vv[0]; v1  = vv[1]; v2  = vv[2]; v3  = vv[3];
			v4  = vv[4]; v5  = vv[5]; v6  = vv[6];

			t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
			t4 += v4 * x4; t5 += v5 * x5; t6 += v6 * x6;

			jj += 7; vv += 7;
		}
		for(j-=6;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4+t5+t6;
	}
}

void lis_matvec_crs_u8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5,t6,t7;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj   = A->index;
	vv   = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = t5 = t6 = t7 = 0.0;
		for(j=js+7;j<je;j+=8)
		{
			j0  = jj[0]; j1  = jj[1]; j2  = jj[2]; j3  = jj[3];
			j4  = jj[4]; j5  = jj[5]; j6  = jj[6]; j7  = jj[7];

			x0  = x[j0]; x1  = x[j1]; x2  = x[j2]; x3  = x[j3];
			x4  = x[j4]; x5  = x[j5]; x6  = x[j6]; x7  = x[j7];

			v0  = vv[0]; v1  = vv[1]; v2  = vv[2]; v3  = vv[3];
			v4  = vv[4]; v5  = vv[5]; v6  = vv[6]; v7  = vv[7];

			t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
			t4 += v4 * x4; t5 += v5 * x5; t6 += v6 * x6; t7 += v7 * x7;

			jj += 8; vv += 8;
		}
		for(j-=7;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4+t5+t6+t7;
	}
}

/************************/
/* Pipeline             */
/************************/
void lis_matvec_crs_p1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t0,t1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = 0;
		if( js<je )
		{
			j0  = jj[0]; v0 = vv[0];
			x0  = x[j0];
			for(j=js+1;j<je;j++)
			{
				j1  = jj[1];
				x1  = x[j1];
				v1  = vv[1];
				t0 += v0 * x0;
				jj++;
				x0  = x1;
				v0  = v1;
				vv++;
			}
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0;
	}
}

void lis_matvec_crs_p2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2;
	LIS_SCALAR	t0,t1,t2;
	LIS_SCALAR	x0,x1,x2;
	LIS_SCALAR	v0,v1,v2;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = 0;
		if( je-js<3 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+2;j<je;j+=2)
			{
				j1  = jj[1]; v1 = vv[1]; x1  = x[j1];
				t0 += v0 * x0;

				j2  = jj[2]; v2 = vv[2]; x2  = x[j2];
				t1 += v1 * x1;

				x0  = x2; v0  = v2;
				jj += 2; vv += 2;

			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 1;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1;
	}
}

void lis_matvec_crs_p3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	t0,t1,t2,t3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = 0;
		if( je-js<4 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+3;j<je;j+=3)
			{
				j1  = jj[1]; j2  = jj[2]; j3  = jj[3];
				v1  = vv[1]; v2  = vv[2]; v3  = vv[3];
				x1  = x[j1]; x2  = x[j2]; x3  = x[j3];
				t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2;

				x0  = x3; v0  = v3;
				jj += 3; vv += 3;
			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 2;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2;
	}
}


void lis_matvec_crs_p4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	t0,t1,t2,t3,t4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = 0;
		if( je-js<5 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+4;j<je;j+=4)
			{
				j1  = jj[1]; j2  = jj[2]; j3  = jj[3]; j4  = jj[4];
				v1  = vv[1]; v2  = vv[2]; v3  = vv[3]; v4  = vv[4];
				x1  = x[j1]; x2  = x[j2]; x3  = x[j3]; x4  = x[j4];
				t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;

				x0  = x4; v0  = v4;
				jj += 4; vv += 4;
			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 3;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3;
	}
}

void lis_matvec_crs_p5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = 0;
		if( je-js<6 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+5;j<je;j+=5)
			{
				j1  = jj[1]; j2  = jj[2]; j3  = jj[3]; j4  = jj[4];
				j5  = jj[5];
				v1  = vv[1]; v2  = vv[2]; v3  = vv[3]; v4  = vv[4];
				v5  = vv[5];
				x1  = x[j1]; x2  = x[j2]; x3  = x[j3]; x4  = x[j4];
				x5  = x[j5];
				t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
				t4 += v4 * x4;

				x0  = x5; v0  = v5;
				jj += 5; vv += 5;
			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 4;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4;
	}
}

void lis_matvec_crs_p6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5,t6;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = t5 = 0;
		if( je-js<7 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+6;j<je;j+=6)
			{
				j1  = jj[1]; j2  = jj[2]; j3  = jj[3]; j4  = jj[4];
				j5  = jj[5]; j6  = jj[6];
				v1  = vv[1]; v2  = vv[2]; v3  = vv[3]; v4  = vv[4];
				v5  = vv[5]; v6  = vv[6];
				x1  = x[j1]; x2  = x[j2]; x3  = x[j3]; x4  = x[j4];
				x5  = x[j5]; x6  = x[j6];
				t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
				t4 += v4 * x4; t5 += v5 * x5;

				x0  = x6; v0  = v6;
				jj += 6; vv += 6;
			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 5;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4+t5;
	}
}

void lis_matvec_crs_p7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5,t6,t7;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = t5 = t6 = 0;
		if( je-js<8 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+7;j<je;j+=7)
			{
				j1  = jj[1]; j2  = jj[2]; j3  = jj[3]; j4  = jj[4];
				j5  = jj[5]; j6  = jj[6]; j7  = jj[7];
				v1  = vv[1]; v2  = vv[2]; v3  = vv[3]; v4  = vv[4];
				v5  = vv[5]; v6  = vv[6]; v7  = vv[7];
				x1  = x[j1]; x2  = x[j2]; x3  = x[j3]; x4  = x[j4];
				x5  = x[j5]; x6  = x[j6]; x7  = x[j7];
				t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
				t4 += v4 * x4; t5 += v5 * x5; t6 += v6 * x6;

				x0  = x7; v0  = v7;
				jj += 7; vv += 7;
			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 6;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4+t5+t6;
	}
}

void lis_matvec_crs_p8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	t0,t1,t2,t3,t4,t5,t6,t7,t8;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = t5 = t6 = t7 = 0;
		if( je-js<9 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			for(j=js+8;j<je;j+=8)
			{
				j1  = jj[1]; j2  = jj[2]; j3  = jj[3]; j4  = jj[4];
				j5  = jj[5]; j6  = jj[6]; j7  = jj[7]; j8  = jj[8];
				v1  = vv[1]; v2  = vv[2]; v3  = vv[3]; v4  = vv[4];
				v5  = vv[5]; v6  = vv[6]; v7  = vv[7]; v8  = vv[8];
				x1  = x[j1]; x2  = x[j2]; x3  = x[j3]; x4  = x[j4];
				x5  = x[j5]; x6  = x[j6]; x7  = x[j7]; x8  = x[j8];
				t0 += v0 * x0; t1 += v1 * x1; t2 += v2 * x2; t3 += v3 * x3;
				t4 += v4 * x4; t5 += v5 * x5; t6 += v6 * x6; t7 += v7 * x7;

				x0  = x8; v0  = v8;
				jj += 8; vv += 8;
			}
			t0 += v0 * x0;
			jj++; vv++;
			j -= 7;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4+t5+t6+t7;
	}
}

void lis_matvec_crs_p3s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	t0,t1,t2,t3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	y0,y1,y2,y3;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = 0;
		if( je-js<4 )
		{
			j = js;
		}
		else
		{
			j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
			j1  = jj[1]; v1 = vv[1]; x1  = x[j1];
			y0  = v0 * x0;
			for(j=js+4;j<je;j+=3)
			{
				j2  = jj[2]; v2 = vv[2]; x2  = x[j2];
				y1  = v1 * x1;
				t0 += y0;

				jj += 3; vv += 3;

				j0  = jj[0]; v0 = vv[0]; x0  = x[j0];
				y2  = v2 * x2;
				t1 += y1;

				j1  = jj[1]; v1 = vv[1]; x1  = x[j1];
				y0  = v0 * x0;
				t2 += y2;

			}
			y1  = v1 * x1;
			t0 += y0;
			t1 += y1;
			jj += 2; vv += 2;
			j -= 2;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2;
	}
}

void lis_matvec_crs_p5s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,js,je;
	int n;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	t0,t1,t2,t3,t4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	y0,y1,y2,y3,y4;
	LIS_SCALAR	*vv;

	n    = A->n;
	jj = A->index;
	vv = A->value;
	for(i=0; i<n; i++)
	{
		js = A->ptr[i];
		je = A->ptr[i+1];
		t0 = t1 = t2 = t3 = t4 = 0;
		if( je-js<3 )
		{
			j = js;
		}
		else
		{
			j0  = *(jj++); v0 = *(vv++); x0  = x[j0];
			j1  = *(jj++); v1 = *(vv++); x1  = x[j1];
			j2  = *(jj++); v2 = *(vv++); x2  = x[j2];
			y0  = v0 * x0;
			y1  = v1 * x1;
			for(j=js+7;j<je;j+=5)
			{
				j3  = *(jj++); v3 = *(vv++); x3  = x[j3];
				y2  = v2 * x2;
				t0 += y0;

				j4  = *(jj++); v4 = *(vv++); x4  = x[j4];
				y3  = v3 * x3;
				t1 += y1;

				j0  = *(jj++); v0 = *(vv++); x0  = x[j0];
				y4  = v4 * x4;
				t2 += y2;

				j1  = *(jj++); v1 = *(vv++); x1  = x[j1];
				y0  = v0 * x0;
				t3 += y3;

				j2  = *(jj++); v2 = *(vv++); x2  = x[j2];
				y1  = v1 * x1;
				t4 += y4;

			}
			y2  = v2 * x2;
			t0 += y0;
			t1 += y1;
			t2 += y2;
			j -= 4;
		}
		for(;j<je;j++)
		{
			j0  = jj[0]; x0  = x[j0]; v0  = vv[0];
			t0 += v0 * x0;
			jj++; vv++;
		}
		y[i] = t0+t1+t2+t3+t4;
	}
}

