/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#if defined(USE_SSE2)
#include <emmintrin.h>

void lis_matvec_dia_s4_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	__m128d		x1,x2;
	__m128d		y1,y2;
	__m128d		v1,v2;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			/*
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			yy = &y[js];
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			*/
			js = k;
			je = _min(n,k+h);
			kk = j*n;
			yy = &y[js];
			yb = yy;
			xx = &x[js];
			vv = &A->value[kk+js];

			if( je-js<5 )
			{
				i = js;
			}
			else
			{
				x1  = _mm_load_pd(xx);
				y1  = _mm_load_pd(yy);
				v1  = _mm_load_pd(vv);
				for(i=js+4;i<je;i+=4)
				{
					xx += 2; yy += 2; vv += 2;
					x2  = _mm_load_pd(xx);
					y2  = _mm_load_pd(yy);
					v2  = _mm_load_pd(vv);
					x1  = _mm_mul_pd(x1,v1);
					y1  = _mm_add_pd(y1,x1);
					_mm_store_pd(yb,y1);
					yb += 2;

					xx += 2; yy += 2; vv += 2;
					x1  = _mm_load_pd(xx);
					y1  = _mm_load_pd(yy);
					v1  = _mm_load_pd(vv);
					x2  = _mm_mul_pd(x2,v2);
					y2  = _mm_add_pd(y2,x2);
					_mm_store_pd(yb,y2);
					yb += 2;
				}
				x1  = _mm_mul_pd(x1,v1);
				y1  = _mm_add_pd(y1,x1);
				_mm_store_pd(yb,y1);
				i-=3;
			}
			for(;i<je;i++)
			{ 
				*yy += *vv * *xx;
				yy++; xx++; vv++;
			}
		}
	}
}

void lis_matvec_dia_u4s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	__m128d		x1,x2;
	__m128d		y1,y2;
	__m128d		v1,v2;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			yb = yy;
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+3;i<je;i+=4)
			{
				x1  = _mm_loadu_pd(xx);
				y1  = _mm_loadu_pd(yy);
				v1  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x1  = _mm_mul_pd(v1,x1);
				y1  = _mm_add_pd(y1,x1);
			
				_mm_storeu_pd(yb,y1);
				yb += 2;

				x2  = _mm_loadu_pd(xx);
				y2  = _mm_loadu_pd(yy);
				v2  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x2  = _mm_mul_pd(v2,x2);
				y2  = _mm_add_pd(y2,x2);
			
				_mm_storeu_pd(yy,y2);
				yb += 2;
			}
			for(i-=3;i<je;i++)
			{
				y[i] += A->value[kk + i] * x[jj+i];
			}
		}
	}
}

void lis_matvec_dia_u6s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	__m128d		x1,x2,x3;
	__m128d		y1,y2,y3;
	__m128d		v1,v2,v3;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			yb = yy;
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+5;i<je;i+=6)
			{
				x1  = _mm_loadu_pd(xx);
				y1  = _mm_loadu_pd(yy);
				v1  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x1  = _mm_mul_pd(v1,x1);
				y1  = _mm_add_pd(y1,x1);
			
				_mm_storeu_pd(yb,y1);
				yb += 2;

				x2  = _mm_loadu_pd(xx);
				y2  = _mm_loadu_pd(yy);
				v2  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x2  = _mm_mul_pd(v2,x2);
				y2  = _mm_add_pd(y2,x2);
			
				_mm_storeu_pd(yb,y2);
				yb += 2;

				x3  = _mm_loadu_pd(xx);
				y3  = _mm_loadu_pd(yy);
				v3  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x3  = _mm_mul_pd(v3,x3);
				y3  = _mm_add_pd(y3,x3);
			
				_mm_storeu_pd(yb,y3);
				yb += 2;
			}
			for(i-=5;i<je;i++)
			{
				y[i] += A->value[kk + i] * x[jj+i];
			}
		}
	}
}

void lis_matvec_dia_u8s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,jj,h,k;
	int n,nnd,kk;
	LIS_SCALAR t;
	__m128d		x1,x2,x3,x4;
	__m128d		y1,y2,y3,y4;
	__m128d		v1,v2,v3,v4;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n    = A->n;
	nnd  = A->nnd;
	h    = A->bnr;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<nnd;j++)
		{
			jj = A->index[j];
			js = _max(k,-jj);
			je = _min(n,k+h);
			je = _min(je,n-jj);
			kk = j*n;
			yy = &y[js];
			yb = yy;
			xx = &x[jj+js];
			vv = &A->value[kk+js];
			for(i=js+7;i<je;i+=8)
			{
				x1  = _mm_loadu_pd(xx);
				y1  = _mm_loadu_pd(yy);
				v1  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x1  = _mm_mul_pd(v1,x1);
				y1  = _mm_add_pd(y1,x1);
			
				_mm_storeu_pd(yb,y1);
				yb += 2;

				x2  = _mm_loadu_pd(xx);
				y2  = _mm_loadu_pd(yy);
				v2  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x2  = _mm_mul_pd(v2,x2);
				y2  = _mm_add_pd(y2,x2);
			
				_mm_storeu_pd(yb,y2);
				yb += 2;

				x3  = _mm_loadu_pd(xx);
				y3  = _mm_loadu_pd(yy);
				v3  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x3  = _mm_mul_pd(v3,x3);
				y3  = _mm_add_pd(y3,x3);
			
				_mm_storeu_pd(yb,y3);
				yb += 2;

				x4  = _mm_loadu_pd(xx);
				y4  = _mm_loadu_pd(yy);
				v4  = _mm_loadu_pd(vv);
				xx += 2; yy += 2; vv += 2;

				x4  = _mm_mul_pd(v4,x4);
				y4  = _mm_add_pd(y4,x4);
			
				_mm_storeu_pd(yb,y4);
				yb += 2;
			}
			for(i-=7;i<je;i++)
			{
				y[i] += A->value[kk + i] * x[jj+i];
			}
		}
	}
}
#endif
