/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#if defined(USE_SSE2)
#include <emmintrin.h>

void lis_matvec_dns_u2_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,h,k;
	int gn,n,kk;
	LIS_SCALAR t;
	__m128d		x1;
	__m128d		y1;
	__m128d		v1;
	LIS_SCALAR *vv,*xx,*yy;

	n    = A->n;
	gn   = A->gn;
	h    = 512;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	k = 0;
	h = n;
//	for(k=0;k<n;k+=h)
	{
		xx = x;
		for(j=0;j<gn;j++)
		{
			x1 = _mm_set1_pd(*xx);
			is = k;
			ie = _min(n,k+h);
			yy = &y[is];
			vv = &A->value[j*n+is];
			for(i=is+1;i<ie;i+=2)
			{
				y1  = _mm_load_pd(yy);
				v1  = _mm_load_pd(vv);

				v1  = _mm_mul_pd(v1,x1);
				y1  = _mm_add_pd(y1,v1);
			
				_mm_store_pd(yy,y1);
				yy += 2;
				vv += 2;
			}
			for(i-=1;i<ie;i++)
			{
				*yy += *vv * *xx;
				yy++;
				vv++;
//				y[i] += A->value[j*n+i] * x[j];
			}
			xx++;
		}
	}
}

void lis_matvec_dns_u4_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,h,k;
	int gn,n,kk;
	LIS_SCALAR t;
	__m128d		x1,x2;
	__m128d		y1,y2;
	__m128d		v1,v2;
	LIS_SCALAR *vv,*xx,*yy;

	n    = A->n;
	gn   = A->gn;
	h    = 512;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	k = 0;
	h = n;
//	for(k=0;k<n;k+=h)
	{
		xx = x;
		for(j=0;j<gn;j++)
		{
			x1 = _mm_set1_pd(*xx);
			x2 = x1;
			is = k;
			ie = _min(n,k+h);
			yy = &y[is];
			vv = &A->value[j*n+is];
			for(i=is+3;i<ie;i+=4)
			{
				y1  = _mm_load_pd(yy);
				v1  = _mm_load_pd(vv);

				v1  = _mm_mul_pd(v1,x1);
				y1  = _mm_add_pd(y1,v1);
			
				_mm_store_pd(yy,y1);
				yy += 2;
				vv += 2;

				y2  = _mm_load_pd(yy);
				v2  = _mm_load_pd(vv);

				v2  = _mm_mul_pd(v2,x2);
				y2  = _mm_add_pd(y2,v2);
			
				_mm_store_pd(yy,y2);
				yy += 2;
				vv += 2;
			}
			for(i-=3;i<ie;i++)
			{
				*yy += *vv * *xx;
				yy++;
				vv++;
//				y[i] += A->value[j*n+i] * x[j];
			}
			xx++;
		}
	}
}

void lis_matvec_dns_u6_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,h,k;
	int gn,n,kk;
	LIS_SCALAR t;
	__m128d		x1,x2,x3;
	__m128d		y1,y2,y3;
	__m128d		v1,v2,v3;
	LIS_SCALAR *vv,*xx,*yy;

	n    = A->n;
	gn   = A->gn;
	h    = 512;
	for(i=0; i<n; i++)
	{
		y[i] = 0.0;
	}
	k = 0;
	h = n;
//	for(k=0;k<n;k+=h)
	{
		xx = x;
		for(j=0;j<gn;j++)
		{
			x1 = _mm_set1_pd(*xx);
			x2 = _mm_set1_pd(*xx);
			x3 = _mm_set1_pd(*xx);
			is = k;
			ie = _min(n,k+h);
			yy = &y[is];
			vv = &A->value[j*n+is];
			for(i=is+5;i<ie;i+=6)
			{
				y1  = _mm_load_pd(yy);
				v1  = _mm_load_pd(vv);

				v1  = _mm_mul_pd(v1,x1);
				y1  = _mm_add_pd(y1,v1);
			
				_mm_store_pd(yy,y1);
				yy += 2;
				vv += 2;

				y2  = _mm_load_pd(yy);
				v2  = _mm_load_pd(vv);

				v2  = _mm_mul_pd(v2,x2);
				y2  = _mm_add_pd(y2,v2);
			
				_mm_stream_pd(yy,y2);
				yy += 2;
				vv += 2;

				y3  = _mm_load_pd(yy);
				v3  = _mm_load_pd(vv);

				v3  = _mm_mul_pd(v3,x3);
				y3  = _mm_add_pd(y3,v3);
			
				_mm_store_pd(yy,y3);
				yy += 2;
				vv += 2;
			}
			for(i-=5;i<ie;i++)
			{
				*yy += *vv * *xx;
				yy++;
				vv++;
//				y[i] += A->value[j*n+i] * x[j];
			}
			xx++;
		}
	}
}
#endif
