/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

LIS_MATVEC_JDS lis_matvec_jds_xxx[5][16] = { 
	{
		lis_matvec_jds_u1 , lis_matvec_jds_u2 ,lis_matvec_jds_u3 ,lis_matvec_jds_u4,
		lis_matvec_jds_u5 , lis_matvec_jds_u6 ,lis_matvec_jds_u7 ,lis_matvec_jds_u8,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
/*		lis_matvec_jds_u9 , lis_matvec_jds_u10,lis_matvec_jds_u11,lis_matvec_jds_u12,
		lis_matvec_jds_u13, lis_matvec_jds_u14,lis_matvec_jds_u15,lis_matvec_jds_u16,*/
	},
	{
		lis_matvec_jds_p1,lis_matvec_jds_p2 ,lis_matvec_jds_p3 ,lis_matvec_jds_p4,
		lis_matvec_jds_p5,lis_matvec_jds_p6 ,lis_matvec_jds_p7 ,lis_matvec_jds_p8,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	},
	{
		lis_matvec_jds_u1s , lis_matvec_jds_u2s ,lis_matvec_jds_u3s ,lis_matvec_jds_u4s,
		lis_matvec_jds_u5s , lis_matvec_jds_u6s ,lis_matvec_jds_u7s ,lis_matvec_jds_u8s,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
/*		lis_matvec_jds_u9s , lis_matvec_jds_u10s,lis_matvec_jds_u11s,lis_matvec_jds_u12s,
		lis_matvec_jds_u13s, lis_matvec_jds_u14s,lis_matvec_jds_u15s,lis_matvec_jds_u16s,*/
	},
	{
		lis_matvec_jds_p1s,lis_matvec_jds_p2s ,lis_matvec_jds_p3s ,lis_matvec_jds_p4s,
		lis_matvec_jds_p5s,lis_matvec_jds_p6s ,lis_matvec_jds_p7s ,lis_matvec_jds_p8s,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	},
#if defined(USE_SSE2)
	{
		lis_matvec_jds_u2s_sse2,lis_matvec_jds_u2s_sse2 ,lis_matvec_jds_u4s_sse2 ,lis_matvec_jds_u4s_sse2,
		lis_matvec_jds_u6s_sse2,lis_matvec_jds_u6s_sse2 ,lis_matvec_jds_u8s_sse2 ,lis_matvec_jds_u8s_sse2,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	}
#else
	{
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
		NULL,NULL ,NULL ,NULL,
	}
#endif
};

/************************/
/* Unrolling            */
/************************/
void lis_matvec_jds_u1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
#pragma vdir nodep
		for(i=is;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+1;i<ie;i+=2)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;

			yy += 2; vv += 2; jj += 2;
		}
		for(i-=1;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1,y2;
	LIS_SCALAR	x0,x1,x2;
	LIS_SCALAR	v0,v1,v2;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+2;i<ie;i+=3)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;

			yy += 3; vv += 3; jj += 3;
		}
		for(i-=2;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	y0,y1,y2,y3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+3;i<ie;i+=4)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;

			yy += 4; vv += 4; jj += 4;
		}
		for(i-=3;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	y0,y1,y2,y3,y4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+4;i<ie;i+=5)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;

			yy += 5; vv += 5; jj += 5;
		}
		for(i-=4;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+5;i<ie;i+=6)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;

			yy += 6; vv += 6; jj += 6;
		}
		for(i-=5;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+6;i<ie;i+=7)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;

			yy += 7; vv += 7; jj += 7;
		}
		for(i-=6;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+7;i<ie;i+=8)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;

			yy += 8; vv += 8; jj += 8;
		}
		for(i-=7;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u9(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+8;i<ie;i+=9)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;

			yy += 9; vv += 9; jj += 9;
		}
		for(i-=8;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u10(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+9;i<ie;i+=10)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;

			yy += 10; vv += 10; jj += 10;
		}
		for(i-=9;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u11(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+10;i<ie;i+=11)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
			j10 = jj[10]; v10 = vv[10]; x10 = x[j10];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;
			yy[10] += v10 * x10;

			yy += 11; vv += 11; jj += 11;
		}
		for(i-=10;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u12(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+11;i<ie;i+=12)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
			j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
			j11 = jj[11]; v11 = vv[11]; x11 = x[j11];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;
			yy[10] += v10 * x10;
			yy[11] += v11 * x11;

			yy += 12; vv += 12; jj += 12;
		}
		for(i-=11;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u13(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+12;i<ie;i+=13)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
			j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
			j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
			j12 = jj[12]; v12 = vv[12]; x12 = x[j12];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;
			yy[10] += v10 * x10;
			yy[11] += v11 * x11;
			yy[12] += v12 * x12;

			yy += 13; vv += 13; jj += 13;
		}
		for(i-=12;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u14(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12,j13;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12,y13;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+13;i<ie;i+=14)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
			j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
			j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
			j12 = jj[12]; v12 = vv[12]; x12 = x[j12];
			j13 = jj[13]; v13 = vv[13]; x13 = x[j13];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;
			yy[10] += v10 * x10;
			yy[11] += v11 * x11;
			yy[12] += v12 * x12;
			yy[13] += v13 * x13;

			yy += 14; vv += 14; jj += 14;
		}
		for(i-=13;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u15(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12,j13,j14;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12,y13,y14;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+14;i<ie;i+=15)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
			j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
			j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
			j12 = jj[12]; v12 = vv[12]; x12 = x[j12];
			j13 = jj[13]; v13 = vv[13]; x13 = x[j13];
			j14 = jj[14]; v14 = vv[14]; x14 = x[j14];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;
			yy[10] += v10 * x10;
			yy[11] += v11 * x11;
			yy[12] += v12 * x12;
			yy[13] += v13 * x13;
			yy[14] += v14 * x14;

			yy += 15; vv += 15; jj += 15;
		}
		for(i-=14;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u16(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12,j13,j14,j15;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12,y13,y14,y15;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14,x15;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14,v15;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		for(i=is+15;i<ie;i+=16)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
			j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
			j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
			j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
			j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
			j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
			j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
			j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
			j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
			j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
			j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
			j12 = jj[12]; v12 = vv[12]; x12 = x[j12];
			j13 = jj[13]; v13 = vv[13]; x13 = x[j13];
			j14 = jj[14]; v14 = vv[14]; x14 = x[j14];
			j15 = jj[15]; v15 = vv[15]; x15 = x[j15];

			yy[0] += v0 * x0;
			yy[1] += v1 * x1;
			yy[2] += v2 * x2;
			yy[3] += v3 * x3;
			yy[4] += v4 * x4;
			yy[5] += v5 * x5;
			yy[6] += v6 * x6;
			yy[7] += v7 * x7;
			yy[8] += v8 * x8;
			yy[9] += v9 * x9;
			yy[10] += v10 * x10;
			yy[11] += v11 * x11;
			yy[12] += v12 * x12;
			yy[13] += v13 * x13;
			yy[14] += v14 * x14;
			yy[15] += v15 * x15;

			yy += 16; vv += 16; jj += 16;
		}
		for(i-=15;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

/************************/
/* Pipeline             */
/************************/
void lis_matvec_jds_p1(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];
		if( is<ie )
		{
			yy  = A->work;
			jj  = &A->index[is];
			vv  = &A->value[is];
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+1;i<ie;i++)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				v0 = v1; x0 = x1;

			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2;
	LIS_SCALAR	y0,y1,y2;
	LIS_SCALAR	x0,x1,x2;
	LIS_SCALAR	v0,v1,v2;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<3 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+2;i<ie;i+=2)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;

				yy += 2; vv += 2; jj += 2;
				v0 = v2; x0 = x2;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=1;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p3(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	y0,y1,y2,y3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<4 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+3;i<ie;i+=3)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;

				yy += 3; vv += 3; jj += 3;
				v0 = v3; x0 = x3;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=2;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p4(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	y0,y1,y2,y3,y4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<5 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+4;i<ie;i+=4)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;

				yy += 4; vv += 4; jj += 4;
				v0 = v4; x0 = x4;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=3;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p5(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<6 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+5;i<ie;i+=5)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;

				yy += 5; vv += 5; jj += 5;
				v0 = v5; x0 = x5;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=4;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p6(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<7 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+6;i<ie;i+=6)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;

				yy += 6; vv += 6; jj += 6;
				v0 = v6; x0 = x6;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=5;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p7(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<8 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+7;i<ie;i+=7)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];  
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;

				yy += 7; vv += 7; jj += 7;
				v0 = v7; x0 = x7;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=6;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p8(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(j=0;j<maxnzr;j++)
	{
		is = A->ptr[j];
		ie = A->ptr[j+1];

		yy  = A->work;
		jj  = &A->index[is];
		vv  = &A->value[is];
		if( ie-is<9 )
		{
			i = is;
		}
		else
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			for(i=is+8;i<ie;i+=8)
			{
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];  
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];  
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];  

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;

				yy += 8; vv += 8; jj += 8;
				v0 = v8; x0 = x8;
			}
			yy[0] += v0 * x0;
			yy++; vv++; jj++;
			i-=7;
		}
		for(;i<ie;i++)
		{
			j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
			yy[0] += v0 * x0;
			yy++; vv++; jj++;

		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

/*************************/
/* Unrolling+Stripmining */
/*************************/
void lis_matvec_jds_u1s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u2s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+1;i<ie;i+=2)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;

				yy += 2; vv += 2; jj += 2;
			}
			for(i-=1;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u3s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1,y2;
	LIS_SCALAR	x0,x1,x2;
	LIS_SCALAR	v0,v1,v2;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+2;i<ie;i+=3)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;

				yy += 3; vv += 3; jj += 3;
			}
			for(i-=2;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u4s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	y0,y1,y2,y3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+3;i<ie;i+=4)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;

				yy += 4; vv += 4; jj += 4;
			}
			for(i-=3;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u5s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	y0,y1,y2,y3,y4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+4;i<ie;i+=5)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;

				yy += 5; vv += 5; jj += 5;
			}
			for(i-=4;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u6s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+5;i<ie;i+=6)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;

				yy += 6; vv += 6; jj += 6;
			}
			for(i-=5;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u7s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+6;i<ie;i+=7)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;

				yy += 7; vv += 7; jj += 7;
			}
			for(i-=6;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u8s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+7;i<ie;i+=8)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;

				yy += 8; vv += 8; jj += 8;
			}
			for(i-=7;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u9s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+8;i<ie;i+=9)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;

				yy += 9; vv += 9; jj += 9;
			}
			for(i-=8;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u10s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+9;i<ie;i+=10)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;

				yy += 10; vv += 10; jj += 10;
			}
			for(i-=9;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u11s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+10;i<ie;i+=11)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
				j10 = jj[10]; v10 = vv[10]; x10 = x[j10];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;
				yy[10] += v10 * x10;

				yy += 11; vv += 11; jj += 11;
			}
			for(i-=10;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u12s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+11;i<ie;i+=12)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
				j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
				j11 = jj[11]; v11 = vv[11]; x11 = x[j11];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;
				yy[10] += v10 * x10;
				yy[11] += v11 * x11;

				yy += 12; vv += 12; jj += 12;
			}
			for(i-=11;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u13s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+12;i<ie;i+=13)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
				j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
				j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
				j12 = jj[12]; v12 = vv[12]; x12 = x[j12];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;
				yy[10] += v10 * x10;
				yy[11] += v11 * x11;
				yy[12] += v12 * x12;

				yy += 13; vv += 13; jj += 13;
			}
			for(i-=12;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u14s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12,j13;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12,y13;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+13;i<ie;i+=14)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
				j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
				j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
				j12 = jj[12]; v12 = vv[12]; x12 = x[j12];
				j13 = jj[13]; v13 = vv[13]; x13 = x[j13];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;
				yy[10] += v10 * x10;
				yy[11] += v11 * x11;
				yy[12] += v12 * x12;
				yy[13] += v13 * x13;

				yy += 14; vv += 14; jj += 14;
			}
			for(i-=13;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u15s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12,j13,j14;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12,y13,y14;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+14;i<ie;i+=15)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
				j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
				j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
				j12 = jj[12]; v12 = vv[12]; x12 = x[j12];
				j13 = jj[13]; v13 = vv[13]; x13 = x[j13];
				j14 = jj[14]; v14 = vv[14]; x14 = x[j14];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;
				yy[10] += v10 * x10;
				yy[11] += v11 * x11;
				yy[12] += v12 * x12;
				yy[13] += v13 * x13;
				yy[14] += v14 * x14;

				yy += 15; vv += 15; jj += 15;
			}
			for(i-=14;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_u16s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8,j9,j10,j11,j12,j13,j14,j15;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8,y9,y10,y11,y12,y13,y14,y15;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12,x13,x14,x15;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8,v9,v10,v11,v12,v13,v14,v15;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+15;i<ie;i+=16)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
				j2 = jj[2]; v2 = vv[2]; x2 = x[j2];
				j3 = jj[3]; v3 = vv[3]; x3 = x[j3];
				j4 = jj[4]; v4 = vv[4]; x4 = x[j4];
				j5 = jj[5]; v5 = vv[5]; x5 = x[j5];
				j6 = jj[6]; v6 = vv[6]; x6 = x[j6];
				j7 = jj[7]; v7 = vv[7]; x7 = x[j7];
				j8 = jj[8]; v8 = vv[8]; x8 = x[j8];
				j9 = jj[9]; v9 = vv[9]; x9 = x[j9];
				j10 = jj[10]; v10 = vv[10]; x10 = x[j10];
				j11 = jj[11]; v11 = vv[11]; x11 = x[j11];
				j12 = jj[12]; v12 = vv[12]; x12 = x[j12];
				j13 = jj[13]; v13 = vv[13]; x13 = x[j13];
				j14 = jj[14]; v14 = vv[14]; x14 = x[j14];
				j15 = jj[15]; v15 = vv[15]; x15 = x[j15];

				yy[0] += v0 * x0;
				yy[1] += v1 * x1;
				yy[2] += v2 * x2;
				yy[3] += v3 * x3;
				yy[4] += v4 * x4;
				yy[5] += v5 * x5;
				yy[6] += v6 * x6;
				yy[7] += v7 * x7;
				yy[8] += v8 * x8;
				yy[9] += v9 * x9;
				yy[10] += v10 * x10;
				yy[11] += v11 * x11;
				yy[12] += v12 * x12;
				yy[13] += v13 * x13;
				yy[14] += v14 * x14;
				yy[15] += v15 * x15;

				yy += 16; vv += 16; jj += 16;
			}
			for(i-=15;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

/************************/
/* Pipeline+Stripmining */
/************************/
void lis_matvec_jds_p1s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;
	LIS_SCALAR	y0,y1;
	LIS_SCALAR	x0,x1;
	LIS_SCALAR	v0,v1;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);
			if( is<ie )
			{
				yy  = &A->work[k];
				jj  = &A->index[is];
				vv  = &A->value[is];
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+1;i<ie;i++)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];  
					yy[0] += v0 * x0;
					yy++; vv++; jj++;
					v0 = v1; x0 = x1;

				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p2s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2;
	LIS_SCALAR	y0,y1,y2;
	LIS_SCALAR	x0,x1,x2;
	LIS_SCALAR	v0,v1,v2;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<3 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+2;i<ie;i+=2)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;

					yy += 2; vv += 2; jj += 2;
					v0 = v2; x0 = x2;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=1;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p3s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	y0,y1,y2,y3;
	LIS_SCALAR	x0,x1,x2,x3;
	LIS_SCALAR	v0,v1,v2,v3;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<4 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+3;i<ie;i+=3)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;

					yy += 3; vv += 3; jj += 3;
					v0 = v3; x0 = x3;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=2;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p4s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4;
	LIS_SCALAR	y0,y1,y2,y3,y4;
	LIS_SCALAR	x0,x1,x2,x3,x4;
	LIS_SCALAR	v0,v1,v2,v3,v4;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<5 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+4;i<ie;i+=4)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
					j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;
					yy[3] += v3 * x3;

					yy += 4; vv += 4; jj += 4;
					v0 = v4; x0 = x4;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=3;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p5s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<6 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+5;i<ie;i+=5)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
					j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
					j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;
					yy[3] += v3 * x3;
					yy[4] += v4 * x4;

					yy += 5; vv += 5; jj += 5;
					v0 = v5; x0 = x5;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=4;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p6s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<7 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+6;i<ie;i+=6)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
					j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
					j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  
					j6 = jj[6]; v6 = vv[6]; x6 = x[j6];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;
					yy[3] += v3 * x3;
					yy[4] += v4 * x4;
					yy[5] += v5 * x5;

					yy += 6; vv += 6; jj += 6;
					v0 = v6; x0 = x6;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=5;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p7s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<8 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+7;i<ie;i+=7)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
					j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
					j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  
					j6 = jj[6]; v6 = vv[6]; x6 = x[j6];  
					j7 = jj[7]; v7 = vv[7]; x7 = x[j7];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;
					yy[3] += v3 * x3;
					yy[4] += v4 * x4;
					yy[5] += v5 * x5;
					yy[6] += v6 * x6;

					yy += 7; vv += 7; jj += 7;
					v0 = v7; x0 = x7;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=6;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

void lis_matvec_jds_p8s(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,k,is,ie,h;
	int n,maxnzr;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7,j8;
	LIS_SCALAR	y0,y1,y2,y3,y4,y5,y6,y7,y8;
	LIS_SCALAR	x0,x1,x2,x3,x4,x5,x6,x7,x8;
	LIS_SCALAR	v0,v1,v2,v3,v4,v5,v6,v7,v8;
	LIS_SCALAR	t;
	LIS_SCALAR	*vv,*xx,*yy;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<9 )
			{
				i = is;
			}
			else
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				for(i=is+8;i<ie;i+=8)
				{
					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
					j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  
					j5 = jj[5]; v5 = vv[5]; x5 = x[j5];  
					j6 = jj[6]; v6 = vv[6]; x6 = x[j6];  
					j7 = jj[7]; v7 = vv[7]; x7 = x[j7];  
					j8 = jj[8]; v8 = vv[8]; x8 = x[j8];  

					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;
					yy[3] += v3 * x3;
					yy[4] += v4 * x4;
					yy[5] += v5 * x5;
					yy[6] += v6 * x6;
					yy[7] += v7 * x7;

					yy += 8; vv += 8; jj += 8;
					v0 = v8; x0 = x8;
				}
				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=7;
			}
			for(;i<ie;i++)
			{
				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  
				yy[0] += v0 * x0;
				yy++; vv++; jj++;

			}
		}
	}
	yy = A->work;
	for(i=0;i<n;i++)
	{
		y[A->row[i]] = yy[i];
	}
}

