/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#if defined(USE_SSE2)
#include <emmintrin.h>

void lis_matvec_jds_u2s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,h,k;
	int n,maxnzr,kk;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	xt0,xt1,xt2,xt3;
	LIS_SCALAR	vt0;
	__m128d		x0,x2;
	__m128d		y0,y2;
	__m128d		v0,v2;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			yb  = yy;
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+1;i<ie;i+=2)
			{
				j0 = jj[0]; xt0 = x[j0];
				j1 = jj[1]; xt1 = x[j1];
				x0  = _mm_set_pd(xt1,xt0);
				y0  = _mm_load_pd(yy);
				v0  = _mm_load_pd(vv);

				yy += 2; vv += 2;
				x0  = _mm_mul_pd(x0,v0);
				y0  = _mm_add_pd(y0,x0);
				_mm_store_pd(yb,y0);
				yb += 2;
			}
			for(i-=1;i<je;i++)
			{ 
				j0 = jj[0]; vt0 = vv[0]; xt0 = x[j0];  
				yy[0] += vt0 * xt0;
				yy++; vv++; jj++;
			}
		}
	}
}

void lis_matvec_jds_u4s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,h,k;
	int n,maxnzr,kk;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	xt0,xt1,xt2,xt3;
	LIS_SCALAR	vt0;
	__m128d		x0,x2;
	__m128d		y0,y2;
	__m128d		v0,v2;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			yb  = yy;
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+3;i<ie;i+=4)
			{
				j0 = jj[0]; xt0 = x[j0];
				j1 = jj[1]; xt1 = x[j1];
				j2 = jj[2]; xt2 = x[j2];
				j3 = jj[3]; xt3 = x[j3];
				x0  = _mm_set_pd(xt1,xt0);
				y0  = _mm_load_pd(yy);
				v0  = _mm_load_pd(vv);
				x2  = _mm_set_pd(xt3,xt2);
				y2  = _mm_load_pd(&yy[2]);
				v2  = _mm_load_pd(&vv[2]);

				yy += 4; vv += 4;
				x0  = _mm_mul_pd(x0,v0);
				y0  = _mm_add_pd(y0,x0);
				x2  = _mm_mul_pd(x2,v2);
				y2  = _mm_add_pd(y2,x2);
				_mm_store_pd(yb,y0);
				_mm_store_pd(&yb[2],y2);
				yb += 4;
			}
			for(i-=3;i<je;i++)
			{ 
				j0 = jj[0]; vt0 = vv[0]; xt0 = x[j0];  
				yy[0] += vt0 * xt0;
				yy++; vv++; jj++;
			}
		}
	}
}

void lis_matvec_jds_u6s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,h,k;
	int n,maxnzr,kk;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5;
	LIS_SCALAR	xt0,xt1,xt2,xt3,xt4,xt5;
	LIS_SCALAR	vt0;
	__m128d		x0,x2,x4;
	__m128d		y0,y2,y4;
	__m128d		v0,v2,v4;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			yb  = yy;
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+5;i<ie;i+=6)
			{
				j0 = jj[0]; xt0 = x[j0];
				j1 = jj[1]; xt1 = x[j1];
				j2 = jj[2]; xt2 = x[j2];
				j3 = jj[3]; xt3 = x[j3];
				j4 = jj[4]; xt4 = x[j4];
				j5 = jj[5]; xt5 = x[j5];
				x0  = _mm_set_pd(xt1,xt0);
				y0  = _mm_load_pd(yy);
				v0  = _mm_load_pd(vv);
				x2  = _mm_set_pd(xt3,xt2);
				y2  = _mm_load_pd(&yy[2]);
				v2  = _mm_load_pd(&vv[2]);
				x4  = _mm_set_pd(xt5,xt4);
				y4  = _mm_load_pd(&yy[4]);
				v4  = _mm_load_pd(&vv[4]);

				yy += 6; vv += 6;
				x0  = _mm_mul_pd(x0,v0);
				y0  = _mm_add_pd(y0,x0);
				x2  = _mm_mul_pd(x2,v2);
				y2  = _mm_add_pd(y2,x2);
				x4  = _mm_mul_pd(x4,v4);
				y4  = _mm_add_pd(y4,x4);
				_mm_store_pd(yb,y0);
				_mm_store_pd(&yb[2],y2);
				_mm_store_pd(&yb[4],y4);
				yb += 6;
			}
			for(i-=5;i<je;i++)
			{ 
				j0 = jj[0]; vt0 = vv[0]; xt0 = x[j0];  
				yy[0] += vt0 * xt0;
				yy++; vv++; jj++;
			}
		}
	}
}

void lis_matvec_jds_u8s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,h,k;
	int n,maxnzr,kk;
	int			*jj;
	int			j0,j1,j2,j3,j4,j5,j6,j7;
	LIS_SCALAR	xt0,xt1,xt2,xt3,xt4,xt5,xt6,xt7;
	LIS_SCALAR	vt0;
	__m128d		x0,x2,x4,x6;
	__m128d		y0,y2,y4,y6;
	__m128d		v0,v2,v4,v6;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			yb  = yy;
			jj  = &A->index[is];
			vv  = &A->value[is];
			for(i=is+7;i<ie;i+=8)
			{
				j0 = jj[0]; xt0 = x[j0];
				j1 = jj[1]; xt1 = x[j1];
				j2 = jj[2]; xt2 = x[j2];
				j3 = jj[3]; xt3 = x[j3];
				j4 = jj[4]; xt4 = x[j4];
				j5 = jj[5]; xt5 = x[j5];
				j6 = jj[6]; xt6 = x[j6];
				j7 = jj[7]; xt7 = x[j7];
				x0  = _mm_set_pd(xt1,xt0);
				y0  = _mm_load_pd(yy);
				v0  = _mm_load_pd(vv);
				x2  = _mm_set_pd(xt3,xt2);
				y2  = _mm_load_pd(&yy[2]);
				v2  = _mm_load_pd(&vv[2]);
				x4  = _mm_set_pd(xt5,xt4);
				y4  = _mm_load_pd(&yy[4]);
				v4  = _mm_load_pd(&vv[4]);
				x6  = _mm_set_pd(xt7,xt6);
				y6  = _mm_load_pd(&yy[6]);
				v6  = _mm_load_pd(&vv[6]);

				yy += 8; vv += 8;
				x0  = _mm_mul_pd(x0,v0);
				y0  = _mm_add_pd(y0,x0);
				x2  = _mm_mul_pd(x2,v2);
				y2  = _mm_add_pd(y2,x2);
				x4  = _mm_mul_pd(x4,v4);
				y4  = _mm_add_pd(y4,x4);
				x6  = _mm_mul_pd(x6,v6);
				y6  = _mm_add_pd(y6,x6);
				_mm_store_pd(yb,y0);
				_mm_store_pd(&yb[2],y2);
				_mm_store_pd(&yb[4],y4);
				_mm_store_pd(&yb[6],y6);
				yb += 8;
			}
			for(i-=7;i<je;i++)
			{ 
				j0 = jj[0]; vt0 = vv[0]; xt0 = x[j0];  
				yy[0] += vt0 * xt0;
				yy++; vv++; jj++;
			}
		}
	}
}

void lis_matvec_jds_p2s_sse2(LIS_MATRIX A, LIS_SCALAR x[], LIS_SCALAR y[])
{
	int i,j,is,ie,js,je,h,k;
	int n,maxnzr,kk;
	int			*jj;
	int			j0,j1,j2,j3;
	LIS_SCALAR	xt0,xt1,xt2,xt3;
	LIS_SCALAR	vt0;
	__m128d		x0,x2;
	__m128d		y0,y2;
	__m128d		v0,v2;
	LIS_SCALAR *vv,*xx,*yy,*yb;

	n      = A->n;
	maxnzr = A->maxnzr;
	yy     = A->work;
	h      = A->bnr;
	for(i=0; i<n; i++)
	{
		yy[i] = 0.0;
	}
	for(k=0;k<n;k+=h)
	{
		for(j=0;j<maxnzr;j++)
		{
			is = A->ptr[j]+k;
			ie = _min(is+h,A->ptr[j+1]);

			yy  = &A->work[k];
			yb  = yy;
			jj  = &A->index[is];
			vv  = &A->value[is];
			if( ie-is<2 )
			{
				i = is;
			}
			else
			{
/*				j0 = jj[0]; v0 = vv[0]; x0 = x[j0];  */
				j0 = jj[0]; xt0 = x[j0];
				j1 = jj[1]; xt1 = x[j1];
				x0  = _mm_set_pd(xt1,xt0);
				y0  = _mm_load_pd(yy);
				v0  = _mm_load_pd(vv);
				for(i=is+3;i<ie;i+=2)
				{
/*					j1 = jj[1]; v1 = vv[1]; x1 = x[j1];
					j2 = jj[2]; v2 = vv[2]; x2 = x[j2];  
					j3 = jj[3]; v3 = vv[3]; x3 = x[j3];  
					j4 = jj[4]; v4 = vv[4]; x4 = x[j4];  */
					yy += 2; vv += 2;
					j2 = jj[2]; xt2 = x[j2];
					j3 = jj[3]; xt3 = x[j3];
					x2  = _mm_set_pd(xt3,xt2);
					y2  = _mm_load_pd(yy);
					v2  = _mm_load_pd(vv);

/*					yy[0] += v0 * x0;
					yy[1] += v1 * x1;
					yy[2] += v2 * x2;
					yy[3] += v3 * x3;*/
					x0  = _mm_mul_pd(x0,v0);
					y0  = _mm_add_pd(y0,x0);
					_mm_store_pd(yb,y0);
					yb += 2;

/*					yy += 4; vv += 4; jj += 4;
					v0 = v4; x0 = x4;*/
					y0 = y2; v0 = v2; x0 = x2;
				}
/*				yy[0] += v0 * x0;
				yy++; vv++; jj++;
				i-=3;*/
				x0  = _mm_mul_pd(x0,v0);
				y0  = _mm_add_pd(y0,x0);
				_mm_store_pd(yb,y0);
				i-=1;
			}
			for(;i<je;i++)
			{ 
				j0 = jj[0]; vt0 = vv[0]; xt0 = x[j0];  
				yy[0] += vt0 * xt0;
				yy++; vv++; jj++;
			}
		}
	}
}

#endif
