/* Copyright (C) 2004-2005 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"
#if defined(USE_SSE2)
#include <emmintrin.h>

int lis_vector_nrm2_u2_sse2(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	__m128d		t1;
	__m128d		r1;
	__m128d		s1;
	LIS_SCALAR d[2];
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	s1   = _mm_setzero_pd();
	for(i=1; i<n; i+=2)
	{
		t1  = _mm_load_pd(x);
		x  += 2;
		r1  = t1;

		t1  = _mm_mul_pd(t1,r1);
		s1  = _mm_add_pd(s1,t1);
	}
	_mm_store_pd(d,s1);
	dot = d[0]+d[1];
	for(i-=1;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_u4_sse2(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,k,n;
	LIS_SCALAR dot,t;
	__m128d		t1,t2;
	__m128d		r1,r2;
	__m128d		s1,s2;
	LIS_SCALAR d[2];
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

#define CSIZE 512
	s1   = _mm_setzero_pd();
	s2   = _mm_setzero_pd();
	for(k=0;k<n;k+=CSIZE)
	{
//		_mm_prefetch((char *)&x[k+CSIZE], _MM_HINT_NTA);
		
		for(i=k;i<k+CSIZE;i+=128)
		{
			_mm_prefetch((char *)&x[i], _MM_HINT_NTA);
		}
		
		for(i=k;i<k+CSIZE;i+=4)
		{
//			_mm_prefetch((char *)(x+4), _MM_HINT_NTA);
			t1  = _mm_load_pd(x);
			x  += 2;
			r1  = t1;

			t2  = _mm_load_pd(x);
			x  += 2;
			r2  = t2;

			t1  = _mm_mul_pd(t1,r1);
			t2  = _mm_mul_pd(t2,r2);
			s1  = _mm_add_pd(s1,t1);
			s2  = _mm_add_pd(s2,t2);
		}
	}
	s1  = _mm_add_pd(s1,s2);
	_mm_store_pd(d,s1);
	dot = d[0]+d[1];
	for(i-=3;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_u6_sse2(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	__m128d		t1,t2,t3;
	__m128d		r1,r2,r3;
	__m128d		s1,s2,s3;
	LIS_SCALAR d[2];
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	s1   = _mm_setzero_pd();
	s2   = _mm_setzero_pd();
	s3   = _mm_setzero_pd();
	for(i=5; i<n; i+=6)
	{
		t1  = _mm_load_pd(x);
		x  += 2;
		r1  = t1;

		t2  = _mm_load_pd(x);
		x  += 2;
		r2  = t2;

		t3  = _mm_load_pd(x);
		x  += 2;
		r3  = t3;

		t1  = _mm_mul_pd(t1,r1);
		t2  = _mm_mul_pd(t2,r2);
		t3  = _mm_mul_pd(t3,r3);
		s1  = _mm_add_pd(s1,t1);
		s2  = _mm_add_pd(s2,t2);
		s3  = _mm_add_pd(s3,t3);
	}
	s1  = _mm_add_pd(s1,s2);
	s1  = _mm_add_pd(s1,s3);
	_mm_store_pd(d,s1);
	dot = d[0]+d[1];
	for(i-=5;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}

int lis_vector_nrm2_s4_sse2(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n;
	LIS_SCALAR dot,t;
	__m128d		t1,t2;
	__m128d		r1,r2;
	__m128d		s1,s2;
	LIS_SCALAR d[2];
	LIS_SCALAR *x;

	x    = vx->value;
	n    = vx->n;

	s1   = _mm_setzero_pd();
	s2   = _mm_setzero_pd();
	t1  = _mm_load_pd(x);
	x  += 2;
	for(i=5; i<n; i+=4)
	{
		r1  = t1;
		t1  = _mm_mul_pd(t1,r1);
		s1  = _mm_add_pd(s1,t1);

		t2  = _mm_load_pd(x);
		x  += 2;
		r2  = t2;
		t2  = _mm_mul_pd(t2,r2);
		s2  = _mm_add_pd(s2,t2);

		t1  = _mm_load_pd(x);
		x  += 2;
	}
	r1  = t1;
	t1  = _mm_mul_pd(t1,r1);
	s1  = _mm_add_pd(s1,t1);
	s1  = _mm_add_pd(s1,s2);
	_mm_store_pd(d,s1);
	dot = d[0]+d[1];
	for(i-=3;i<n;i++)
	{
		t    = *(x++);
		dot += t*t;
	}
	*val = sqrt(dot);
	return LIS_SUCCESS;
}
#endif
